#include <vector>
#include <map>
#include <string>

class cJavaClass;
class cJVM;
struct METHOD_INFO;
struct FIELD_INFO;
struct STACK_ITEM;

//-----------------------------------------------------------------------------
// Name: class cVirtualMachine
// Desc: Reprezents virtual machine.
//-----------------------------------------------------------------------------
class cJVM
{
private:
	//typedef stl map for easier use
	typedef std::map<std::string,cJavaClass*>				ClassMap;
	typedef std::map<std::string,cJavaClass*>::iterator		ClassMapItor;
	typedef std::map<std::string,cJavaClass*>::value_type	ClassMapValue;

	//map of all loaded classes
	ClassMap	m_Classes;

	//Array of pointers to instruction functions
	void (cJVM::*Instructions[INST_NUMINST])();

	//program counter
	u1	*m_pPC;
	//stack
	cStack			m_Stack;
	//heap
	cHeap			m_Heap;
	//class of current method
	cJavaClass		*m_pCurClass;
	//flag tells if an error occured
	bool			m_bError;

private:
	//Read u2 (big endian)
	u2 ReadU2();
	//Read u4 (big endian)
	u4 ReadU4();
	//Get double from high, low bytes
	void GetDouble(jDouble &d,u4 iHighBytes,u4 iLowBytes);
	//Get long from high, low bytes
	void GetLong(jLong &l,u4 iHighBytes,u4 iLowBytes);
	//Get high, low bytes from double
	void GetBytesFromDouble(jDouble &d,u4 &iHighBytes,u4 &iLowBytes);
	//Get high, low bytes from long
	void GetBytesFromLong(jLong &l,u4 &iHighBytes,u4 &iLowBytes);
	//create recursively multi dimensional array
	OBJECT* CreateMultiDimArray(u1 iCurDim,u1 iNumDim,u1 iItemSize,cJavaClass *pItem);
	//check cast
	bool CheckCast(OBJECT *pObj,u2 iConstantIndex);
	//check cast
	bool CheckCastRecursive(OBJECT *pObj,cJavaClass *pClass,u1 iDimensions,u1 iItemSize);

	//Breaks execution and prints error msg to stdout
	void Error(char *strError);

	//Instructions
	//new
	void Inst_New();
	//new array
	void Inst_NewArray();
	//anew array
	void Inst_ANewArray();
	//multi anew array
	void Inst_MultiANewArray();
	//check cast
	void Inst_CheckCast();
	//instance of
	void Inst_InstanceOf();
	//wide
	void Inst_Wide();
	//dup
	void Inst_Dup();
	//dup x1
	void Inst_Dup_X1();
	//dup x2
	void Inst_Dup_X2();
	//dup2
	void Inst_Dup2();
	//dup2 x1
	void Inst_Dup2_X1();
	//dup2 x2
	void Inst_Dup2_X2();
	//invoke interface
	void Inst_InvokeInterface();
	//invoke virtual
	void Inst_InvokeVirtual();
	//invoke special
	void Inst_InvokeSpecial();
	//invoke static
	void Inst_InvokeStatic();
	//ret
	void Inst_Ret();
	//aload
	void Inst_ALoad();
	//lload
	void Inst_LLoad();
	//lload 0
	void Inst_LLoad_0();
	//lload 1
	void Inst_LLoad_1();
	//lload 2
	void Inst_LLoad_2();
	//lload 3
	void Inst_LLoad_3();
	//dload
	void Inst_DLoad();
	//dload 0
	void Inst_DLoad_0();
	//dload 1
	void Inst_DLoad_1();
	//dload 2
	void Inst_DLoad_2();
	//dload 3
	void Inst_DLoad_3();
	//aload 0
	void Inst_ALoad_0();
	//aload 1
	void Inst_ALoad_1();
	//aload 2
	void Inst_ALoad_2();
	//aload 3
	void Inst_ALoad_3();
	//return
	void Inst_Return();
	//ireturn
	void Inst_IReturn();
	//freturn
	void Inst_FReturn();
	//areturn
	void Inst_AReturn();
	//dreturn
	void Inst_DReturn();
	//lreturn
	void Inst_LReturn();
	//dstore
	void Inst_DStore();
	//lstore
	void Inst_LStore();
	//astore
	void Inst_AStore();
	//fstore
	void Inst_FStore();
	//fstore 0
	void Inst_FStore_0();
	//fstore 1
	void Inst_FStore_1();
	//fstore 2
	void Inst_FStore_2();
	//fstore 3
	void Inst_FStore_3();
	//astore 0
	void Inst_AStore_0();
	//astore 1
	void Inst_AStore_1();
	//astore 2
	void Inst_AStore_2();
	//astore 3
	void Inst_AStore_3();
	//istore
	void Inst_IStore();
	//istore 1
	void Inst_IStore_1();
	//istore 0
	void Inst_IStore_0();
	//istore 2
	void Inst_IStore_2();
	//istore 3
	void Inst_IStore_3();
	//dstore 0
	void Inst_DStore_0();
	//dstore 1
	void Inst_DStore_1();
	//dstore 2
	void Inst_DStore_2();
	//dstore 3
	void Inst_DStore_3();
	//lstore 0
	void Inst_LStore_0();
	//lstore 1
	void Inst_LStore_1();
	//lstore 2
	void Inst_LStore_2();
	//lstore 3
	void Inst_LStore_3();
	//iconst m1
	void Inst_IConst_M1();
	//iconst 0
	void Inst_IConst_0();
	//iconst 1
	void Inst_IConst_1();
	//iconst 2
	void Inst_IConst_2();
	//iconst 3
	void Inst_IConst_3();
	//iconst 4
	void Inst_IConst_4();
	//iconst 5
	void Inst_IConst_5();
	//fload
	void Inst_FLoad();
	//fload 0
	void Inst_FLoad_0();
	//fload 1
	void Inst_FLoad_1();
	//fload 2
	void Inst_FLoad_2();
	//fload 3
	void Inst_FLoad_3();
	//iload
	void Inst_ILoad();
	//iload 0
	void Inst_ILoad_0();
	//iload 1
	void Inst_ILoad_1();
	//iload 2
	void Inst_ILoad_2();
	//iload 3
	void Inst_ILoad_3();
	//iadd
	void Inst_IAdd();
	//iand
	void Inst_IAnd();
	//idiv
	void Inst_IDiv();
	//ishl
	void Inst_ISHL();
	//ishr
	void Inst_ISHR();
	//iushr
	void Inst_IUSHR();
	//irem
	void Inst_IRem();
	//ior
	void Inst_IOr();
	//ixor
	void Inst_IXOr();
	//ineg
	void Inst_INeg();
	//imul
	void Inst_IMul();
	//isub
	void Inst_ISub();
	//fadd
	void Inst_FAdd();
	//fsub
	void Inst_FSub();
	//fdiv
	void Inst_FDiv();
	//fmul
	void Inst_FMul();
	//frem
	void Inst_FRem();
	//fneg
	void Inst_FNeg();
	//pop
	void Inst_Pop();
	//pop2
	void Inst_Pop2();
	//iastore
	void Inst_IAStore();
	//sastore
	void Inst_SAStore();
	//fastore
	void Inst_FAStore();
	//dastore
	void Inst_DAStore();
	//lastore
	void Inst_LAStore();
	//bipush
	void Inst_BIPush();
	//put static
	void Inst_PutStatic();
	//get static
	void Inst_GetStatic();
	//put field
	void Inst_PutField();
	//get field
	void Inst_GetField();
	//goto
	void Inst_GoTo();
	//goto w
	void Inst_GoTo_W();
	//baload
	void Inst_BALoad();
	//caload
	void Inst_CALoad();
	//iinc
	void Inst_IInc();
	//array length
	void Inst_ArrayLength();
	//if i cmp eq
	void Inst_If_ICMPEQ();
	//if i cmp ne
	void Inst_If_ICMPNE();
	//if i cmp lt
	void Inst_If_ICMPLT();
	//if i cmp le
	void Inst_If_ICMPLE();
	//if i cmp gt
	void Inst_If_ICMPGT();
	//if i cmp ge
	void Inst_If_ICMPGE();
	//if a cmp eq
	void Inst_If_ACMPEQ();
	//if a cmp ne
	void Inst_If_ACMPNE();
	//if eq
	void Inst_IFEQ();
	//if ne
	void Inst_IFNE();
	//if lt
	void Inst_IFLT();
	//if le
	void Inst_IFLE();
	//if gt
	void Inst_IFGT();
	//if ge
	void Inst_IFGE();
	//if non null
	void Inst_IfNonNull();
	//if non null
	void Inst_IfNull();
	//sipush
	void Inst_SIPush();
	//bastore
	void Inst_BAStore();
	//castore
	void Inst_CAStore();
	//iaload
	void Inst_IALoad();
	//saload
	void Inst_SALoad();
	//iaload
	void Inst_FALoad();
	//daload
	void Inst_DALoad();
	//aaload
	void Inst_AALoad();
	//ldc
	void Inst_LDC();
	//ldc w
	void Inst_LDC_W();
	//ldc2 w
	void Inst_LDC2_W();
	//dadd
	void Inst_DAdd();
	//dsub
	void Inst_DSub();
	//drem
	void Inst_DRem();
	//dneg
	void Inst_DNeg();
	//dmul
	void Inst_DMul();
	//ddiv
	void Inst_DDiv();
	//d2i
	void Inst_D2I();
	//d2f
	void Inst_D2F();
	//d2l
	void Inst_D2L();
	//l2d
	void Inst_L2D();
	//l2f
	void Inst_L2F();
	//l2i
	void Inst_L2I();
	//laload
	void Inst_LALoad();
	//ladd
	void Inst_LAdd();
	//ldiv
	void Inst_LDiv();
	//lsub
	void Inst_LSub();
	//lshl
	void Inst_LShl();
	//lshr
	void Inst_LShr();
	//lushr
	void Inst_LUShr();
	//lor
	void Inst_LOr();
	//lxor
	void Inst_LXOr();
	//lrem
	void Inst_LRem();
	//lneg
	void Inst_LNeg();
	//lmul
	void Inst_LMul();
	//land
	void Inst_LAnd();
	//lcmp
	void Inst_LCmp();
	//dconst 0
	void Inst_DConst_0();
	//dconst 1
	void Inst_DConst_1();
	//lconst 0
	void Inst_LConst_0();
	//lconst 1
	void Inst_LConst_1();
	//aastore
	void Inst_AAStore();
	//aconst null
	void Inst_AConst_Null();
	//dcmpl
	void Inst_DCMPL();
	//dcmpg
	void Inst_DCMPG();
	//fcmpl
	void Inst_FCMPL();
	//fcmpg
	void Inst_FCMPG();
	//f2d
	void Inst_F2D();
	//f2l
	void Inst_F2L();
	//f2i
	void Inst_F2I();
	//fconst 0
	void Inst_FConst_0();
	//fconst 1
	void Inst_FConst_1();
	//fconst 2
	void Inst_FConst_2();
	//i2b
	void Inst_I2B();
	//i2c
	void Inst_I2C();
	//i2d
	void Inst_I2D();
	//i2f
	void Inst_I2F();
	//i2l
	void Inst_I2L();
	//i2s
	void Inst_I2S();
	//jsr
	void Inst_JSR();
	//jsr w
	void Inst_JSR_W();
	//swap
	void Inst_Swap();
	//nop
	void Inst_Nop() {}

	//relocate single object (garbage collector)
	OBJECT* CollectObject(OBJECT *pObj);

public:
	//Constructor
	cJVM();
	//Destructor
	~cJVM();
	
	//Removes all loaded classes
	void FreeAllClasses();
	//Initialize vm. Must be called before anything else
	bool Initialize(u4 iStackSize, u4 iHeapSize);
	//Load class from file and prepare its main method for execution
	bool LoadMain(const char *strFilename);
	//Load class from file
	cJavaClass* LoadClass(const char *strFilename);
	//Run initialization method of suplied class. Called automaticly after class is loaded.
	bool InitializeClass(cJavaClass *pClass);
	//Run method if any was prepared for execution
	bool Run();
	//Run garbage collector
	void RunGC();
};